from copy import deepcopy
import selection_strategy as ss
import robustness
from robustness import model_utils, datasets, \
    defaults, imagenet_models, cifar_models,  \
    train, data_augmentation as da
from robustness.datasets import DATASETS
import numpy as np
from sklearn.decomposition import PCA as SKPCA
from tqdm import tqdm
from typing import List, Dict, Tuple
from cox.utils import Parameters
import cox.store
import torch
from torchvision import transforms
import torchvision

'''
These functions are intended to be combined with strategies in
selection_strategy.py to produce selection functions.
Major methods in this module:

input_space_selector 
projected_space_selector 
random_selector
feature_space_selector

In this entire module, selection_func is applied sequentially to each class, so
you may provide a general callable that changes state with each call as well,
and it will work
'''

def _adjacency_list_from_edges(distance_mask, num_nodes: int):
    '''
    Internal. Produces an adjacency list
    representation from edges of a graph.
    '''
    adjacency_list = [[] for _ in range(num_nodes)]
    for u, v in distance_mask:
        adjacency_list[u.item()].append(v.item())
    return adjacency_list
    

def _process_batch(
    batch: torch.tensor, 
    points: torch.tensor, 
    norm: float, 
    norm_threshold: float):
    '''
    Internal. Takes a batch of vectors, 
    and a large batch of other vectors,
    and produces an adjacency list
    of points within a certain threshold.
    '''
    if torch.cuda.is_available():
        batch = batch.to('cuda')
        points = points.to('cuda')
    distance_matrix = torch.cdist(x1=batch.unsqueeze(0), x2=points.unsqueeze(0), p=norm).squeeze()
    distance_mask = distance_matrix <= norm_threshold
    # distance_mask = torch.nonzero(distance_mask)
    adjacency_list = [torch.nonzero(i).view(-1).cpu().numpy() for i in distance_mask]
    # adjacency_list = _adjacency_list_from_edges(distance_mask, batch.size(0))
    return adjacency_list


def _remove_self_loops(adjacency_list: List[List[int]]) -> List[List[int]]:
    '''
    Internal. Removes self loops from a given
    adjacency list
    '''
    return [np.setdiff1d(l, [i], True) for i, l in enumerate(adjacency_list)]


def get_connections(
    points, 
    norm: float, 
    norm_threshold: float, 
    batch_size: int, 
    desc: str = None, 
    use_tqdm: bool = False) -> List[List[int]]:
    '''
    Returns an adjacency list over
    points' indices built over the input 'norm'
    and 'norm_threshold'
    Batched processing for finding pairwise
    distances
    '''
    if(len(points) == 0):
        return []
    points = np.asarray(points)
    points = points.reshape([len(points), -1])
    points = torch.tensor(points)
    points_loader = torch.utils.data.DataLoader(points, batch_size=batch_size)
    if use_tqdm:
        points_loader = tqdm(points_loader, desc=desc)
    adjacency_list = []
    for batch in points_loader:
        batch_adj = _process_batch(batch, points, norm, norm_threshold)
        adjacency_list.extend(batch_adj)
    adjacency_list = _remove_self_loops(adjacency_list)
    return adjacency_list



# Selection strategies

def input_space_selector(points, norm, 
    norm_threshold: float, 
    selection_algorithm, 
    batch_size: int, 
    budgets: List[int],
    desc: str = None) -> Dict[int, List[bool]]:
    '''
    Returns a mask 
    selection_algorithm: should be a strategy like those in selection_strategy.py
    '''
    points = np.asarray(points).reshape([len(points), -1])
    n_vertices = len(points)
    adjacency_list = get_connections(
        points=points,
        norm=norm,
        norm_threshold=norm_threshold,
        batch_size=batch_size,
        desc=desc
    )
    return selection_algorithm(adjacency_list, budgets)


def projected_space_selector(
    points, 
    projected_dim, 
    norm, 
    norm_threshold: float, 
    selection_algorithm, 
    batch_size: int, 
    budgets: List[int]) -> Dict[int, List[bool]]:

    pca = SKPCA(n_components=projected_dim)

    points = np.asarray(points).reshape([len(points), -1])
    projected_points = pca.fit_transform(points)
    return input_space_selector(
        points=points,
        norm=norm,
        norm_threshold=norm_threshold,
        selection_algorithm=selection_algorithm,
        batch_size=batch_size,
        budgets=budgets
    )


def _batched_feature_extractor(points, feature_extractor, use_tqdm=False):
    projected_points = []
    batch_size = 128
    loader = torch.utils.data.DataLoader(points, batch_size=batch_size, shuffle=False)
    if use_tqdm:
        loader = tqdm(loader)
    for batch in loader:
        output = feature_extractor(batch)
        if type(output) != np.ndarray:
            output = output.cpu().detach().numpy()
        projected_points.append(output)
    projected_points = np.concatenate(projected_points)
    return projected_points


def feature_space_selector(
    points,
    feature_extractor,
    norm,
    norm_threshold: float,
    selection_algorithm,
    batch_size: int,
    budgets: List[int]) -> Dict[int, List[bool]]:

    points = np.asarray(points).reshape([len(points), -1])
    projected_points = _batched_feature_extractor(points, feature_extractor)
    return input_space_selector(
        points=projected_points,
        norm=norm,
        norm_threshold=norm_threshold,
        selection_algorithm=selection_algorithm,
        batch_size=batch_size,
        budgets=budgets
    )


def _get_random_selection(points, budget: int):
    '''
    Internal. Helper for random selector.
    '''
    chosen_indices = np.arange(len(points))
    np.random.shuffle(chosen_indices)
    chosen_indices = chosen_indices[: budget]
    chosen_mask = [False for _ in points]
    for chosen_index in chosen_indices:
        chosen_mask[chosen_index] = True
    return chosen_mask


def random_selector(
    points: List[List[int]], 
    budgets: List[int]) -> Dict[int, List[bool]]:
    '''
    randomly selects a mask, unirformly
    '''
    result_dict = {}
    for budget in budgets:
        result_dict[budget] = _get_random_selection(points, budget)
    return result_dict


def get_poisoned_labels(
    labels: List[int], 
    mask: List[bool], 
    num_classes: int) -> List[int]:
    '''
    Returns a new labels list with values moved to 
    the next label where mask is True
    '''
    labels = [(label + int(mask_val)) % num_classes \
        for label, mask_val in zip(labels, mask)]
    return labels

def get_poison_mask(x: np.ndarray, y: List[int],
    selection_func) -> List[bool]:
    '''
    Takes in the data points and labels,
    returns a boolean mask over this set
    marking labels to flip
    Parameters:
    x: datapoints,
    y: labels,
    selection_func: a callable mapping classpoints
        to a bool mask. All points from the
        same class will be plugged into this 
        function to decide which points to flip.
    '''
    assert x.shape[0] == len(y)
    # assert len(x.shape) == 2, 'x has shape {}'.format(x.shape)
    num_classes = len(set(y))
    classwise_points = [[] for _ in range(num_classes)]
    for point, label in zip(x, y):
        classwise_points[label].append(point)
    classwise_flip_masks = [selection_func(class_points) for \
        class_points in classwise_points]
    counters = [0 for _ in range(num_classes)]
    result = []
    for point, label in zip(x, y):
        result.append(classwise_flip_masks[label][counters[label]])
        counters[label] += 1
    return result

def unzip(x):
    '''
    Takes a list x = [(x1, y1), (x2, y2), ...]
    Returns [[x1, x2, ...], [y1, y2, ...]]
    '''
    return list(zip(*x))

def _get_split_mask(mask, y):
    num_classes = len(set(y))
    result = [[] for _ in range(num_classes)]
    for mask_val, y_val in zip(mask, y):
        result[y_val].append(mask_val)
    return result

def get_poisoned_dataset(dataset, selection_func, transform, return_dataset: bool):
    '''
    Takes in a dataset, applies selection function
    to create a flipped dataset, and returns a new
    "poisoned" dataset
    '''
    x, y = unzip(dataset)
    numpy_x = list(map(np.asarray, x))
    numpy_x = np.asarray(numpy_x).astype(np.float32)
    numpy_x = numpy_x / 255
    num_classes = len(set(y))
    poison_mask = get_poison_mask(
        x=numpy_x.reshape([len(numpy_x), -1]), 
        y=y, 
        selection_func=selection_func
    )
    flipped_y = get_poisoned_labels(labels=y, mask=poison_mask, \
        num_classes=num_classes)
    if return_dataset:
        return TransformedDataSet(list(zip(x, flipped_y)), transform)
    else:
        return _get_split_mask(poison_mask, y)

class TransformedDataSet(torch.utils.data.Dataset):
    def __init__(self,
        dataset,
        transform):
        self.transform = transform
        self.dataset = dataset
    
    def __len__(self):
        return len(self.dataset)
    
    def __getitem__(self, idx):
        x, y = self.dataset[idx]
        x = self.transform(x)
        return (x, y)


if __name__ == '__main__':
    import os
    HOME = os.environ['HOME']
    selection_strategy = ss.greedy
    selector = random_selector
    def selection_func(points):
        return [False for _ in points]
    ds = CIFAR(
        data_path=os.path.join(HOME, 'datasets/cifar'),
    )
    model, checkpoint_loaded = robustness.model_utils.make_and_restore_model(
        arch='resnet18',
        dataset=ds,
        resume_path=None,
    )

    train_loader, val_loader = ds.make_loaders(batch_size=128, \
            shuffle=True, workers=2, selection_func=selection_func)

    # Create a cox store for logging
    out_store = cox.store.Store(os.path.join(HOME, 'logs'))

    # Hard-coded base parameters
    train_kwargs = {
        'out_dir': os.path.join(HOME, 'models'),
        'adv_train': 0,
        'epochs': 100
    }
    train_args = Parameters(train_kwargs)

    # Fill whatever parameters are missing from the defaults
    train_args = defaults.check_and_fill_args(train_args,
                            defaults.TRAINING_ARGS, DATASETS['cifar'])
    # train_args = defaults.check_and_fill_args(train_args,
    #                         defaults.PGD_ARGS, CIFAR)
    
    # Train a model
    train.train_model(train_args, model, (train_loader, val_loader), \
        store=out_store, checkpoint=checkpoint_loaded)


def _get_resnet18_pca_feature_extractor(model_path: str, data_path: str, projected_dim: int):
    dataset = datasets.CIFAR(data_path)
    model, _ = model_utils.make_and_restore_model(arch='resnet18', dataset=dataset)
    print('model loaded')
    model = model.model
    model = model.children()
    model = list(model)[: 4]
    model = torch.nn.Sequential(*model)
    # model = torch.nn.Sequential(da.TEST_TRANSFORMS_DEFAULT, *model, torch.nn.Flatten())
    # initial_layers = lambda x: torch.nn.Flatten()(model(da.TEST_TRANSFORMS_DEFAULT(x)))
    def initial_layers(x):
        x = torch.reshape(x, [x.size(0), 3, 32, 32])
        output = model(x.cuda())
        output = torch.nn.Flatten()(output)
        return output
    dataset = unzip(torchvision.datasets.CIFAR10(data_path, train=True))[0]
    dataset = list(map(da.TEST_TRANSFORMS_DEFAULT(32), dataset))

    print('Extracting features', end='')
    dataset = _batched_feature_extractor(dataset, initial_layers, use_tqdm=True)
    print('Features extracted')
    from sklearn.decomposition import PCA
    print('Fitting PCA', end='')
    pca = PCA(n_components=projected_dim)
    pca.fit(dataset)
    print('PCA fit')
    feature_extractor = lambda x: pca.transform(initial_layers(x).cpu().detach().numpy())
    return feature_extractor

